library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

-- interface do ArbitroBarramento
entity ArbitroBarramento is
	port(
	clock:		in std_logic;
	reset:		in std_logic;
	request:	in std_logic_vector(2 downto 0);
	release:	in std_logic_vector(2 downto 0);
	busy:		out std_logic;
	abort:		out std_logic;
	grant:		out std_logic_vector(2 downto 0));
end ArbitroBarramento;

-- implementao do ArbitroBarramento
architecture ArbitroBarramento of ArbitroBarramento is

type state is (S0,S1,S2,S3);
signal ES, PES: state;
signal pede: std_logic;
signal counter,timeout: std_logic_vector(5 downto 0);
signal sel,prox: integer; 
begin

	timeout <= "110010"; -- timeout igual a 50 ciclos ou 1000ns.
	
	pede <= '1' when request(0)='1' or request(1)='1' or request(2)='1' else
		'0';
	
		
	process(reset,clock)
	begin
		if reset='1' then
			ES<=S0;
		elsif clock'event and clock='1' then
			ES<=PES;
		end if;
	end process;

	process(sel,request,prox)
	begin
		case sel is
			when 0=> 
				if request(1)='1' then
					prox<=1;
				elsif request(2)='1' then
					prox<=2;
				elsif request(0)='1' then
					prox<=0;
				else
					prox<=1;
				end if;
			when 1=> 
				if request(2)='1' then
					prox<=2;
				elsif request(0)='1' then
					prox<=0;
				elsif request(1)='1' then
					prox<=1;
				else
					prox<=2;
				end if;
			when 2=> 
				if request(0)='1' then
					prox<=0;
				elsif request(1)='1' then
					prox<=1;
				elsif request(2)='1' then
					prox<=2;
				else
					prox<=0;
				end if;
			when others => prox<=0;
		end case;
	end process;		
			
	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR o prximo estado da mquina.
	------------------------------------------------------------------------------------------------------	
	--S0 -> Arbitro verifica se request(0)=1 se for avanca para S1 senao
	--		avanca para S4.
	--S1 -> Arbitro recebeu Request(0), ativa o Busy e o Grant(0).
	--S2 -> Desativa o Grant(0) e dispara o contador de timeout. Se Release(0) for 1 entao
	--		avanca para S4, senao se contador igual a timeout avanca para S3.
	--S3 -> Eh ativado o Abort indicando timeout.
	--others    -> Retorna a S0.
	process(ES,pede,release,sel,counter,timeout)
	begin
		case ES is
			-- core 0
			when S0 => if pede='1' then PES <= S1; else PES <= S0; end if;
			when S1 => PES <= S2;
			when S2 => if release(sel)='1' then PES <= S0;
					   elsif counter=timeout then PES<= S3;
					   else PES <= S2;
					   end if;
			when S3 => PES <= S0;
		end case;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados
	------------------------------------------------------------------------------------------------------	
	process (clock)
	begin
		if clock'event and clock='0' then
			case ES is
				-- Verifica se recebeu request do core 0
				when S0 =>
					busy <= '0';
					abort <= '0';
					grant <= (others => '0');
					counter <= (others => '0');
					sel <= prox;
				-- Recebeu request do core 0
				when S1 =>
					grant(sel) <= '1';
					busy <= '1';
					
				-- Espera recepcao do release ou que expire o timeout
				when S2 =>
					grant(sel) <= '0';
					counter <= counter + '1';
					
				-- Ativa o sinal de abort
				when S3 =>
					abort <= '1';
			end case;
		end if;
	end process;	

end ArbitroBarramento;
