 ---------------------------------------------------------------------------------------	
--                                SENDRECEIVE
---------------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use work.process_packet.all;

-- interface do Send e do Receive
entity SendReceive is
	port(
	clock: 		in std_logic;
	reset: 		in std_logic;
	addressC:	in std_logic_vector(7 downto 0);
	avSend:		in std_logic;
	nwordSend:	in std_logic_vector(2 downto 0);
	busySend:	out std_logic;
	busyWrapper:in std_logic;
	dataSend:	in packet;
	dataReceive:out packet;
	avReceive:	out std_logic;
	busy: 		in std_logic;
	abort: 		in std_logic;
	request: 	out std_logic;
	grant: 		in std_logic;
	release: 	out std_logic;
	data_in:	in std_logic_vector(15 downto 0);
	data_out:	out std_logic_vector(15 downto 0));
end SendReceive;

-- implementao do SendReceive
architecture SendReceive of SendReceive is

type stateSendWrapper is (SW0,SW1,SW2,SW3,SW4,SW5,SW6);
signal SW, PSW: stateSendWrapper;

type stateSendAck is (SA0,SA1,SA2,SA3,SA4,SA5);
signal SA, PSA: stateSendAck;

type stateReceive is (SR0,SR1,SR2,SR3,SR4,SR5,SR6,SR7);
signal SR, PSR: stateReceive;

signal ack, avWrapper, avAck, sendWrapper, sendAck: std_logic;
signal nwordS,nwordAckR,nwordAck,nwordWrapper: std_logic_vector(2 downto 0);
signal counterTimeout,timeout: std_logic_vector(19 downto 0);
signal counterRepeat,nrepeat: std_logic_vector(3 downto 0);
signal counterWordAck,counterWordWrapper,counterWordReceive: integer;
signal dataS,dataR,dataWrapper,dataAckR,dataAck: packet;

begin
	avWrapper <= avSend;
	--retirar daqui
	timeout <= x"3FFFF";	
	--ate aqui para simular
	--retirar daqui
	--timeout <= x"0000F";
	--ate aqui para prototipar
	nrepeat <= x"F";
	
	process(reset,clock)
	begin
		if reset='1' then
			SW<=SW0;
			SA<=SA0;
			SR<=SR0;
		elsif clock'event and clock='1' then
			SW<=PSW;
			SA<=PSA;
			SR<=PSR;
		end if;
	end process;

	
-----------------------------------------------------------------------------------------------------	
-- SEND
------------------------------------------------------------------------------------------------------	
	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR O PRXIMO ESTADO DA MQUINA DO SEND WRAPPER
	------------------------------------------------------------------------------------------------------	
	process(SW,SA,avWrapper,grant,counterWordWrapper,nwordS,counterTimeout,timeout,counterRepeat,nrepeat,ack)
	begin
		case SW is
			when SW0 => if avWrapper ='1' then PSW <= SW1; else PSW <= SW0; end if;
			when SW1 => PSW <= SW2;
			when SW2 => if SA=SA0 then PSW <= SW3; else PSW <= SW2; end if;
			when SW3 => if counterRepeat/=x"0" and ack ='1' then PSW <= SW0;
						elsif grant='1' then PSW <= SW4;
						else PSW <= SW3; end if;
			when SW4 => if ack ='1' then PSW <= SW0;
						elsif counterWordWrapper=nwordS then PSW <= SW5;
						else PSW <= SW4; end if;
			when SW5 => if ack ='1' or counterRepeat=nrepeat then PSW <= SW0;
						else PSW <= SW6; end if;
			when SW6 => if ack='1' then PSW <= SW0;
						elsif counterTimeout=timeout then PSW <= SW2;
						else PSW <= SW6; end if;
		end case;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados SEND WRAPPER
	------------------------------------------------------------------------------------------------------	
	process (clock)
	begin
		if clock'event and clock='0' then
			case SW is
				when SW0 =>
					sendWrapper <= '0';
					counterWordWrapper <= 0;
					counterRepeat <= (others => '0');
					counterTimeout <= (others => '0');
				when SW1 =>
					nwordWrapper <= nwordSend;
					dataWrapper<= dataSend;
				when SW2 =>
					counterWordWrapper <= 0;
				when SW3 =>
					sendWrapper <= '1';
				when SW4 =>
					counterWordWrapper <= counterWordWrapper + 1;
				when SW5 =>
					sendWrapper <= '0';
					counterRepeat <= counterRepeat + '1';
					counterTimeout <= (others=>'0');
				when SW6 =>
					counterTimeout <= counterTimeout + '1';
			end case;
		end if;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR O PRXIMO ESTADO DA MQUINA DO SEND ACK
	------------------------------------------------------------------------------------------------------	
	process(SA,avAck,sendWrapper,grant,counterWordAck,nwordS)
	begin
		case SA is
			when SA0 => if avAck ='1' then PSA <= SA1; else PSA <= SA0; end if;
			when SA1 => PSA <= SA2;
			when SA2 => if sendWrapper='0' then PSA <= SA3; else PSA <= SA2; end if;
			when SA3 => if grant='1' then PSA <= SA4; else PSA <= SA3; end if;
			when SA4 => if counterWordAck=nwordS then PSA <= SA5; else PSA <= SA4; end if;
			when SA5 => PSA <= SA0;
		end case;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados SEND WRAPPER
	------------------------------------------------------------------------------------------------------	
	process (clock)
	begin
		if clock'event and clock='0' then
			case SA is
				when SA0 =>
					sendAck <= '0';
					counterWordAck <= 0;
				-- captura dados do ack
				when SA1 =>
					dataAck <= dataAckR;
					nwordAck<= nwordAckR;
				-- espera o wrapper liberar o send				
				when SA2 =>
					sendAck <= '0';
				-- espera o grant
				when SA3 =>
					sendAck <= '1';
				-- enviando ack
				when SA4 =>
					counterWordAck <= counterWordAck + 1;
				-- ativa release
				when SA5 =>
					sendAck <= '0';
			end case;
		end if;
	end process;
	
	busySend <= '0' when SW=SW0 and SA=SA0 else '1';
	request<='1' when SW=SW3 or SA=SA3 else '0';
	release<='1' when SW=SW5 or SA=SA5 else '0';
		
	process(reset,clock,SW,SA)
	begin
		if reset='1' then
			nwordS<=(others=>'0');
			dataS<=(others=>(others=>'0'));
			data_out<=(others=>'1');
		elsif clock'event and clock='0' then
			if SW=SW3 then
				nwordS <= nwordWrapper;
				dataS <= dataWrapper;
			elsif SA=SA3 then
				nwordS <=nwordAck;
				dataS <= dataAck;
			elsif SW=SW4 then
				data_out <= dataS(counterWordWrapper);
			elsif SA=SA4 then
				data_out <= dataS(counterWordAck);
			else
				data_out <= (others=>'1');
			end if;
		end if;
	end process;
	
-----------------------------------------------------------------------------------------------------	
-- RECEIVE
------------------------------------------------------------------------------------------------------	
	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR O PRXIMO ESTADO DA MQUINA
	------------------------------------------------------------------------------------------------------	
	process(SR,busy,dataR,addressC,abort,busyWrapper,sendWrapper)
	begin
		case SR is
			when SR0 => if busy='1' then PSR <= SR1; else PSR <= SR0; end if;
			when SR1 => PSR <= SR2;
			when SR2 => if dataR(0)(15 downto 8)= addressC then PSR <= SR3; else PSR <= SR7; end if;
			when SR3 => if dataR(1)=x"05" then PSR <= SR4;
						elsif busy='0' and busyWrapper='0' and sendWrapper='0' then PSR <= SR6;
						elsif busy='0' and busyWrapper='0' then PSR <= SR5;
						elsif busy='0' or abort='1' then PSR <= SR0;
						else PSR <= SR3; end if;
			-- ACK
			when SR4 => PSR <= SR0;
			-- Espera Se sendWrapper/='0'
			when SR5 => if sendWrapper='0' then PSR <= SR6; else PSR <= SR5; end if;
			-- ENVIO DO ACK
			when SR6 => PSR <= SR0;
			-- OBSTRUCT
			when SR7 => if busy='0' or abort='1' then PSR <= SR0; else PSR <= SR7; end if;
		end case;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados
	------------------------------------------------------------------------------------------------------	
	process(clock)
	begin
		if clock'event and clock='0' then
			case SR is
				when SR0 =>
					counterWordReceive <= 0;
					avAck <= '0';
					avReceive <= '0';
					ack<= '0';
				when SR1 =>
					avReceive <= '0';
				when SR2 =>
					dataR(counterWordReceive) <= data_in;
					counterWordReceive <= counterWordReceive + 1;
				when SR3 =>
					dataR(counterWordReceive) <= data_in;
					counterWordReceive <= counterWordReceive + 1;
				-- CHEGOU ACK
				when SR4 =>
					ack <= '1';
				-- ESPERA PARA ENVIAR ACK
				when SR5 =>
					avAck <= '0';
				-- ENVIA ACK
				when SR6 =>
					avReceive <= '1';
					dataReceive <= dataR;
					avAck <= '1';
					nwordAckR<= "010";
					dataAckR(0)<= dataR(0)(7 downto 0) & dataR(0)(15 downto 8);
					dataAckR(1)<= x"0005";
				-- OBSTRUCT
				when SR7 =>
					avAck <= '0';
			end case;
		end if;
	end process;

end SendReceive;
	

