library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use work.process_packet.all;

-- interface do WrapperSerial
entity WrapperSerial is
port(
	clock:		in std_logic;
	reset:		in std_logic;
	state:		out std_logic_vector(7 downto 0);
	addressC:	in std_logic_vector(7 downto 0);
---- habilitar os sinais txd e rxd  para prototipar ----
 	txd:		in std_logic;
	rxd:		out std_logic;
---- retirar a partir daqui -------------------------------------
--	rx_data:	out std_logic_vector(7 downto 0);  
--	rx_start:	out std_logic;                     
--	rx_busy:	in std_logic;                     	
--	tx_data:	in std_logic_vector(7 downto 0); 
--	tx_av:		in std_logic;                        	
---- at aqui para prototipar -----------------------------------
	request:	out std_logic;
	grant:		in std_logic;
	release:	out std_logic;
	busy:		in std_logic;
	abort:		in std_logic;
	data_out:	out std_logic_vector(15 downto 0);
	data_in:	in std_logic_vector(15 downto 0));
end;

-- implementao do mdulo serial
architecture WrapperSerial of WrapperSerial is    

-- interface do SendReceive
component SendReceive is
port(
	clock: 		in std_logic;
	reset: 		in std_logic;
	addressC:	in std_logic_vector(7 downto 0);
	avSend:		in std_logic;
	nwordSend:	in std_logic_vector(2 downto 0);
	busySend:	out std_logic;
	busyWrapper:	in std_logic;
	dataSend:	in packet;
	dataReceive:	out packet;
	avReceive:	out std_logic;
	busy: 		in std_logic;
	abort: 		in std_logic;
	request: 	out std_logic;
	grant: 		in std_logic;
	release: 	out std_logic;
	data_in:	in std_logic_vector(15 downto 0);
	data_out:	out std_logic_vector(15 downto 0));
end component;

-- habilitar o componente SerialInterface ----
-- interface da serial
component SerialInterface is
port(
    clock:	in std_logic;                        
    reset:	in std_logic;                        
    rx_data:	in std_logic_vector (7 downto 0);  
    rx_start:	in std_logic;                     
    rx_busy:	out std_logic;                     
    rxd:	out std_logic;                         
    txd:	in std_logic;                          
    tx_data:	out std_logic_vector (7 downto 0); 
    tx_av:	out std_logic);
end component;

type state_reqSend is (SS0,SS1,SS2);
signal SS,PSS: state_reqSend;

type state_type is (S0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,R13,R14, --read
W1,W2,W3,W4,W5,W6,W7,W8,W9,--write
H1,H2,H3,H4,--habilita processador
P1,P2,P3,P4,P5,P6,P7,P8,P9,P10,P11,P12,P13,P14,P15,P16,P17,P18,P19,P20,P21,P22, -- comando printf e scanf vindos do processador
SC1,SC2,SC3,SC4,SC5,SC6); -- retorno do scanf 

signal ES,PES: state_type;
-- retira daqui
signal tx_av, rx_busy, rx_start: std_logic;
signal tx_data, rx_data: std_logic_vector(7 downto 0);
-- ate aqui para simular
signal command : std_logic_vector(7 downto 0);
signal addressCReceive,addressCSource,coreReceive: std_logic_vector(7 downto 0);
signal addressBR: std_logic_vector(15 downto 0);
signal dataBR: std_logic_vector(15 downto 0);
signal regdata: std_logic_vector(7 downto 0);
signal counterWordPC,nwordPC: std_logic_vector(7 downto 0);
signal counter: std_logic_vector(21 downto 0);
signal nvezes: std_logic_vector(2 downto 0);
signal reqSend: std_logic;

-- SendReceive
signal avSend, avReceive, busySend,busySerial:	std_logic;
signal nwordSend: std_logic_vector(2 downto 0);
signal dataSend, dataReceive, dataR:		packet;
signal Nclock:				std_logic;

signal stateWS: std_logic_vector(7 downto 0);

-- implementao do WrapperSerial
begin
	Nclock <= clock;
	state <= stateWS;

	-- Retirar daqui
	--interface serial
	SI : SerialInterface
	port map(
	    clock=> clock,
	    reset=> reset,
	    rx_data=> rx_data,
	    rx_start=> rx_start,
	    rx_busy=> rx_busy,
	    rxd=> rxd,
	    txd=> txd,
	    tx_data=> tx_data,
	    tx_av=> tx_av);
	-- ate aqui para simular
	
	SR : SendReceive
	port map(
	clock => Nclock,
	reset => reset,
	addressC => addressC,
	avSend => avSend,
	nwordSend => nwordSend,
	busySend => busySend,
	busyWrapper => busySerial,
	dataSend => dataSend,
	dataReceive => dataReceive,
	avReceive => avReceive,
	busy => busy,
	abort => abort,
	request => request,
	grant => grant,
	release => release,
	data_in => data_in,
	data_out => data_out);


------------------------------------------------------------------------------------------
--  COMUNICAO COM MEIO EXTERNO
------------------------------------------------------------------------------------------
	--registrador que guarda ultimo byte que vem da serial
	process (tx_av, clock, reset)  
	begin
		if reset = '1' then 
			regdata <= (others => '0');
		elsif clock'event and clock = '1' then 
			if tx_av = '1' then 
				regdata <= tx_data;
			end if;
		end if;
	end process;
	
	process(reset,clock)
	begin
		if reset='1' then
			ES<=S0;
			SS<=SS0;
		elsif clock'event and clock='1' then
			ES<=PES;
			SS<=PSS;
		end if;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR O PRXIMO ESTADO DA MQUINA DE ESTADOS DO PEDE SEND
	------------------------------------------------------------------------------------------------------	
	process (SS, reqSend, busySend)
	begin
		case SS is
			when SS0 => if reqSend= '1' then PSS <= SS1; else PSS <= SS0; end if;
			when SS1 => if busySend='0' then PSS <= SS2; else PSS <= SS1; end if;
			when SS2 => PSS <= SS0;
		end case;
	end process;
	
	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados PEDE SEND
	------------------------------------------------------------------------------------------------------	
	process (clock)
	begin
		if clock'event and clock='0' then
			case SS is
				when SS0 =>
					avSend<= '0';
				when SS1 =>
					avSend<= '0';
				when SS2 =>
					avSend <= '1';	
			end case;
		end if;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR O PRXIMO ESTADO DA MQUINA DE ESTADOS PC->SERIAL SERIAL->PC
	------------------------------------------------------------------------------------------------------	
	process (ES,tx_av,tx_data,busySend,avReceive,dataReceive,addressCSource,coreReceive,rx_busy,counter,counterWordPC,nWordPC,nvezes,command)
	begin
		case ES is
			when S0 => if tx_av='1' and tx_data=x"00" then PES<=R1;
				elsif tx_av='1' and tx_data=x"01" then PES<=W1;
				elsif tx_av='1' and tx_data=x"02" then PES<=H1;
				elsif tx_av='1' and tx_data=x"04" then PES<=SC1;
				elsif avReceive='1' and (dataReceive(1)=x"0003" or dataReceive(1)=x"0004" or dataReceive(1)=x"0006") then PES<=P1;
				elsif avReceive='1' and dataReceive(1)=X"0000" then PES<=R8;
				else PES<=S0; end if;
			-- estados da leitura
			when R1 => if tx_av='1' then PES<=R2; else PES<=R1; end if;
			when R2 => if tx_av='1' then PES<=R3; else PES<=R2; end if;
			when R3 => if tx_av='1' then PES<=R4; else PES<=R3; end if;
			when R4 => if tx_av='1' then PES<=R5; else PES<=R4; end if;
			when R5 => PES<=R6;
			when R6 => PES<=R7;
			when R7 =>  PES<=S0;
			-- recebendo o dado lido
			when R8 => PES<=R9;
			when R9 => if rx_busy='0' then PES<=R10; else PES<=R9; end if;
			when R10 => if counter=x"3FFFF" then PES<=R11; else PES<=R10; end if;
--			when R10 => if counter(3 downto 0)=x"F" then PES<=R11; else PES<=R10; end if;
			when R11 => PES<=R12;
			when R12 => if rx_busy='0' then PES<=R13; else PES<=R12; end if;
			when R13 => if counter=x"3FFFF" then PES<=R14; else PES<=R13; end if;
--			when R13 => if counter(3 downto 0)=x"F" then PES<=R14; else PES<=R13; end if;
			when R14 => if counterWordPC=nWordPC then PES<=S0; else PES<=R6; end if;

			-- estados da escrita
			when W1 => if tx_av='1' then PES<=W2; else PES<=W1; end if;
			when W2 => if tx_av='1' then PES<=W3; else PES<=W2; end if;
			when W3 => if tx_av='1' then PES<=W4; else PES<=W3; end if;
			when W4 => if tx_av='1' then PES<=W5; else PES<=W4; end if;
			when W5 => if tx_av='1' then PES<=W6; else PES<=W5; end if;
			when W6 => if tx_av='1' then PES<=W7; else PES<=W6; end if;
			when W7 => PES<=W8;
			when W8 => if counterWordPC=nWordPC then PES<=S0; else PES<=W9; end if;
			when W9 => if tx_av='1' then PES<=W6; else PES<=W9; end if;

			-- estados de habilitao do processador
			when H1=>  if tx_av='1' then PES<=H2; else PES<=H1; end if;
			when H2=>  PES<=H3;
			when H3=>  PES<=H4;
			when H4=>  PES<=S0;
			
			-- comandos (printf e scanf) do processador --
			when P1=>  PES<=P2;
			when P2=>  PES<=P3;
			when P3=>  if rx_busy='0' then PES<=P4; else PES<=P3; end if;
			when P4=>  if counter=x"3FFFF" and nvezes=x"4" then PES<=P5;
						elsif counter=x"3FFFF" then PES<=P2;
						else PES<=P4; end if;
--			when P4=>  if counter(3 downto 0)=x"F" and nvezes=x"4" then PES<=P5;
--						elsif counter(3 downto 0)=x"F" then PES<=P2;
--						else PES<=P4; end if;
			when P5=>  PES<=P6;
			when P6=>  if rx_busy='0' then PES<=P7; else PES<=P6; end if;
			when P7=>  if counter=x"3FFFF" then PES<=P8; else PES<=P7; end if;
--			when P7=>  if counter(3 downto 0)=x"F" then PES<=P8; else PES<=P7; end if;
			when P8=>  PES<=P9;
			when P9=>  if rx_busy='0' then PES<=P10; else PES<=P9; end if;
			when P10=>  if counter=x"3FFFF" and command=x"04" then PES<=S0;
							elsif counter=x"3FFFF" and (command=x"03" or command=x"06") then PES<=P11;
							else PES<=P10; end if;
--			when P10=>  if counter(3 downto 0)=x"F" and command=x"04" then PES<=S0;
--							elsif counter(3 downto 0)=x"F" and (command=x"03" or command=x"06") then PES<=P11;
--							else PES<=P10; end if;
			when P11=>  PES<=P12;
			when P12=>  if rx_busy='0' then PES<=P13; else PES<=P12; end if;
			when P13=>  if counter=x"3FFFF" then PES<=P14; else PES<=P13; end if;
--			when P13=>  if counter(3 downto 0)=x"F" then PES<=P14; else PES<=P13; end if;
			when P14=>  PES<=P15;
			when P15=>  if rx_busy='0' then PES<=P16; else PES<=P15; end if;
			when P16=>  if counter=x"3FFFF" and command=x"06" then PES<= P17;
					elsif counter=x"3FFFF" then PES<=S0; 
					else PES<=P16; end if;
--			when P16=>  if counter(3 downto 0)=x"F" and command=x"06" then PES<= P17;
--					elsif counter(3 downto 0)=x"F" then PES<=S0; 
--					else PES<=P16; end if;
			when P17=>  PES<=P18;
			when P18=>  if rx_busy='0' then PES<=P19; else PES<=P18; end if;
			when P19=>  if counter=x"3FFFF" then PES<=P20; else PES<=P19; end if;
--			when P19=>  if counter(3 downto 0)=x"F" then PES<=P20; else PES<=P19; end if;
			when P20=>  PES<=P21;
			when P21=>  if rx_busy='0' then PES<=P22; else PES<=P21; end if;
			when P22=>  if counter=x"3FFFF" then PES<=S0; else PES<=P22; end if;
--			when P22=>  if counter(3 downto 0)=x"F" then PES<=S0; else PES<=P22; end if;

			--SCANF
			when SC1=>  if tx_av='1' then PES<=SC2; else PES<=SC1; end if;
			when SC2=>  if tx_av='1' then PES<=SC3; else PES<=SC2; end if;
			when SC3=>  if tx_av='1' then PES<=SC4; else PES<=SC3; end if;
			when SC4=>  PES<=SC5;
			when SC5=>  PES<=SC6;
			when SC6=>  PES<=S0;
		end case;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados
	------------------------------------------------------------------------------------------------------	
	process (clock,ES)
	begin
		if clock'event and clock='0' then
			case ES is
				when S0 =>
					busySerial<= '0';
					reqSend<= '0';
					rx_start <= '0';
					stateWS <= x"00";					
				-- requisitando leitura
				when R1 =>
					busySerial<= '1';
					command <= regdata;
					stateWS <= x"A0";					
				when R2 =>
					coreReceive <= regdata;
					stateWS <= x"B0";					
				when R3 =>
					nWordPC <= regdata;
					counterWordPC <= (others=>'0');
					stateWS <= x"C0";					
				when R4 =>
					addressBR(15 downto 8) <= regdata;
					stateWS <= x"D0";					
				when R5 =>
					addressBR(7 downto 0) <= regdata;
					stateWS <= x"E0";					
				when R6 =>
					dataSend(0) <= coreReceive & addressC;
					dataSend(1) <= x"00" & command;
					dataSend(2) <= addressBR;
					nWordSend <= "011";
					reqSend <= '1';
					stateWS <= x"F0";					
				when R7 =>
					reqSend <='0';
					stateWS <= x"A0";					
				-- recebendo dado lido
				when R8 =>
					busySerial<= '1';
					dataR <= dataReceive;
					rx_data <= dataReceive(2)(15 downto 8);
					rx_start <= '1';
					counterWordPC <= counterWordPC + '1';
					stateWS <= x"10";					
				when R9 =>
					rx_start <= '0';
					counter <= (others=>'0');
					stateWS <= x"20";					
				when R10 =>
					counter <= counter + '1';
					stateWS <= x"30";					
				when R11 =>
					rx_data <= dataR(2)(7 downto 0);
					rx_start <= '1';
					stateWS <= x"40";					
				when R12 =>
					rx_start <= '0';
					counter <= (others=>'0');
					stateWS <= x"50";					
				when R13 =>
					counter <= counter + '1';
					stateWS <= x"60";					
				when R14 =>
					addressBR <= addressBR + '1';
					counter <= (others=>'0');
					stateWS <= x"70";					
				
				-- estados da escrita
				when W1 =>
					busySerial<= '1';
					command <= regdata;
					stateWS <= x"01";					
				when W2 =>
					coreReceive <= regdata;
					stateWS <= x"02";					
				when W3 =>
					nWordPC <= regdata;
					counterWordPC <= (others=>'0');
					stateWS <= x"03";					
				when W4 =>
					addressBR(15 downto 8) <= regdata;
					stateWS <= x"04";					
				when W5 =>
					addressBR(7 downto 0) <= regdata;
					stateWS <= x"05";					
				when W6 =>
					dataBR(15 downto 8) <= regdata;
					stateWS <= x"06";					
				when W7 =>
					dataBR(7 downto 0) <= regdata;
					stateWS <= x"07";					
				when W8 =>
					dataSend(0) <= coreReceive & addressC;
					dataSend(1) <= x"00" & command;
					dataSend(2) <= addressBR;
					dataSend(3) <= dataBR;
					nWordSend <= "100";
					addressBR <= addressBR + '1';
					counterWordPC <= counterWordPC + '1';
					reqSend<='1';
					stateWS <= x"08";					
				when W9 =>
					reqSend<='0';
					stateWS <= x"09";					

				-- habilitacao do processador
				when H1 =>
					busySerial<= '1';
					command <= regdata;
					stateWS <= x"0A";					
				when H2 =>
					coreReceive <= regdata;
					stateWS <= x"0B";					
				when H3 =>
					dataSend(0) <= coreReceive & addressC;
					dataSend(1) <= x"00" & command;
					nWordSend <= "010";
					reqSend <= '1';
					stateWS <= x"0C";					
				when H4 =>
					reqSend <= '0';
					stateWS <= x"0D";					

				-- comandos (scanf ou printf) do processador --
				when P1 =>
					busySerial<= '1';
					dataR <= dataReceive;
					addressCReceive <= dataReceive(0)(15 downto 8);
					addressCSource <= dataReceive(0)(7 downto 0);
					command <= dataReceive(1)(7 downto 0);
					dataBR <= dataReceive(2);
					nvezes <= (others=>'0');
				when P2 =>
					rx_data <= x"55";
					rx_start <= '1';
					nvezes <= nvezes + '1';
				when P3 | P6 | P9 | P12 | P15 | P18 | P21=>
					rx_start <= '0';
					counter <= (others=>'0');
				when P4 | P7 | P10 | P13 | P16 | P19 | P22 =>
					counter <= counter + '1';
				when P5 =>
					rx_data <= addressCSource;
					rx_start <= '1';
				when P8 =>
					rx_data <= command;
					rx_start <= '1';
				when P11 =>
					rx_data <= dataBR(15 downto 8);
					rx_start <= '1';
				when P14 =>
					rx_data <= dataBR(7 downto 0);
					rx_start <= '1';
				when P17 =>
					rx_data <= dataR(3)(15 downto 8);
					rx_start <= '1';
				when P20 =>
					rx_data <= dataR(3)(7 downto 0);
					rx_start <= '1';

				--SCANF
				when SC1 =>
					busySerial<= '1';
					command <= regdata;
				when SC2 =>
					addressCReceive <= regdata;
				when SC3 =>
					dataBR(15 downto 8) <= regdata;
				when SC4 =>
					dataBR(7 downto 0) <= regdata;
				when SC5 =>
					dataSend(0) <= coreReceive & addressC;
					dataSend(1) <= x"00" & command;
					dataSend(2) <= dataBR;
					nWordSend <= "011";
					reqSend <= '1';
				when SC6 =>
					reqSend <= '0';
			end case;
		end if;
	end process;

end WrapperSerial;
